/*
**  Copyright (c) 2003,2004 Andrea Luzzardi <scox@gentoo.org>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <shadow.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utmp.h>

#define PAM_SM_AUTH

#include <security/pam_modules.h>
#include <security/_pam_macros.h>

#include "pam.h"
#include "parser.h"
#include "conf.h"
#include "device.h"
#include "auth.h"

char *converse (pam_handle_t *pamh, int style, char *message)
{
	struct pam_conv *c;
	struct pam_message pam_msg;
	struct pam_response *reply = NULL;

	const struct pam_message *msg_ptr;

	union {
		struct pam_conv *c;
	} cnv;

	if (pam_get_item (pamh, PAM_CONV, (const void **)&cnv.c) != PAM_SUCCESS) {
		DEBUGP ("Cannot start conversation\n");
		return NULL;
	}

	pam_msg.msg_style = style;
	pam_msg.msg = message;
	msg_ptr = &pam_msg;

	c = cnv.c;

	c->conv (1, &msg_ptr, &reply, c->appdata_ptr);

	if (reply && style != PAM_PROMPT_ECHO_OFF)
		_pam_drop_reply (reply, 1);

	return reply->resp;
}

FILE *log_init (void)
{
	FILE *f;

	/* no log file, print to stderr */
	if (!*(usb_conf.log_file))
		return stderr;

	/* log_file has been specified, open it and turn debug on */
	
	if (!(f = fopen (usb_conf.log_file, "a")))
		return stderr;
	
	usb_conf.debug = 1;
	
	return f;
}

void log_fini (FILE *f)
{
	if ((usb_conf.debug == 1) && (f != stderr))
		fclose (f);
}

int local_login (pam_handle_t *pamh)
{
	struct utmp *utent = NULL;
	char **local = NULL;
	char *tty = NULL;
	int ret = 1, i = 0;

	union {
		char *str;
	} cnv;

	if (usb_conf.allow_remote == 1)
		return 1;

	if (pam_get_item (pamh, PAM_TTY, (const void **)&cnv.str) != PAM_SUCCESS) {
		DEBUGP ("Cannot retrieve tty name\n");
		return 0;
	}

	tty = cnv.str;

	local = split (usb_conf.local_consoles, ',');

	while (local[i]) {
		if (!strncmp (tty, local[i], strlen(local[i]))) {
			DEBUGP ("Authentication request from local console (%s)\n", tty);
			drop_split (local);
			return 1;
		}

		++i;
	}

	i = 0;
	drop_split (local);

	utmpname (usb_conf.utmp);
	setutent();

	DEBUGP ("Searching the utmp entry for tty %s...\n", tty);

	while ((utent = getutent ()) != NULL) {
		if (!(ret = strcmp (tty, utent->ut_line)))
			break;
	}

	endutent();

	if (ret != 0) {
		DEBUGP ("Cannot retrieve the utmp entry\n");
		return 0;
	}

	DEBUGP ("Authentication request from %s (%s)\n", tty, utent->ut_host);

	if (!strlen(utent->ut_host))
		return 1;

	local = split (usb_conf.local_hosts, ',');

	while (local[i]) {
		if (!strncmp (utent->ut_host, local[i], strlen(local[i]))) {
			drop_split (local);
			return 1;
		}

		++i;
	}

	drop_split (local);

	return 0;
}

PAM_EXTERN
int pam_sm_authenticate (pam_handle_t *pamh, int flags, int argc
			 ,const char **argv)
{
	const char *user = NULL;

	if (pam_get_user (pamh, &user, NULL) != PAM_SUCCESS)
		return PAM_AUTH_ERR;

	/* Do we trust data returned by PAM functions ? */

	if (!user || strlen(user) <= 0)
		return PAM_AUTH_ERR;

	parse_args (argc, argv);
	put_default_values();

	debug_stream = log_init();

	if (!(local_login(pamh))) {
		DEBUGP ("Authentication denied: remote user\n");
		return PAM_AUTH_ERR;
	}

	if (!(check_device())) {
		DEBUGP("Device not valid.\n");
		return PAM_AUTHINFO_UNAVAIL;
	}

	if (!authenticate(pamh, user)) {
		DEBUGP ("Cannot authenticate user \"%s\"\n", user);
		return PAM_AUTH_ERR;
	}

	DEBUGP("Access granted\n");

	log_fini (debug_stream);
	
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_setcred (pam_handle_t *pamh,int flags,int argc, const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

#ifdef PAM_STATIC

struct pam_module _pam_usb_modstruct = {
	"pam_usb",
	pam_sm_authenticate,
	pam_sm_setcred,
	pam_sm_acct_mgmt,
	pam_sm_open_session,
	pam_sm_close_session,
	pam_sm_chauthtok
};

#endif
